@@ -1,14 +1,20 @@
-#Copyright ReportLab Europe Ltd. 2000-2017
+#Copyright ReportLab Europe Ltd. 2000-2019
 #see license.txt for license details
 # $URI:$
-__version__='3.3.0'
+__version__='3.5.34'
 __doc__='''Gazillions of miscellaneous internal utility functions'''
 
-import os, sys, imp, time, types
-from base64 import decodestring as base64_decodestring, encodestring as base64_encodestring
+import os, sys, time, types, datetime, ast
+from functools import reduce as functools_reduce
+literal_eval = ast.literal_eval
+try:
+    from base64 import decodebytes as base64_decodebytes, encodebytes as base64_encodebytes
+except ImportError:
+    from base64 import decodestring as base64_decodebytes, encodestring as base64_encodebytes
 from reportlab import isPy3
 from reportlab.lib.logger import warnOnce
 from reportlab.lib.rltempfile import get_rl_tempfile, get_rl_tempdir, _rl_getuid
+from . rl_safe_eval import rl_safe_exec, rl_safe_eval, safer_globals
 
 try:
     import cPickle as pickle
@@ -121,10 +127,10 @@ if isPy3:
             return str(x).encode(enc)
 
     def encode_label(args):
-        return base64_encodestring(pickle.dumps(args)).strip().decode('latin1')
+        return base64_encodebytes(pickle.dumps(args)).strip().decode('latin1')
 
     def decode_label(label):
-        return pickle.loads(base64_decodestring(label.encode('latin1')))
+        return pickle.loads(base64_decodebytes(label.encode('latin1')))
 
     def rawUnicode(s):
         '''converts first 256 unicodes 1-1'''
@@ -218,10 +224,10 @@ else:
     from string import letters as ascii_letters, uppercase as ascii_uppercase, lowercase as ascii_lowercase
 
     def encode_label(args):
-        return base64_encodestring(pickle.dumps(args)).strip()
+        return base64_encodebytes(pickle.dumps(args)).strip()
 
     def decode_label(label):
-        return pickle.loads(base64_decodestring(label))
+        return pickle.loads(base64_decodebytes(label))
 
     def rawUnicode(s):
         '''converts first 256 unicodes 1-1'''
@@ -232,6 +238,7 @@ else:
         return s.encode('latin1') if isinstance(s,unicode) else s
 
     def rl_exec(obj, G=None, L=None):
+        '''this is unsafe'''
         if G is None:
             frame = sys._getframe(1)
             G = frame.f_globals
@@ -240,7 +247,7 @@ else:
             del frame
         elif L is None:
             L = G
-        exec("""exec obj in G, L""")
+        exec(obj,G, L)
     rl_exec("""def rl_reraise(t, v, b=None):\n\traise t, v, b\n""")
 
     char2int = ord
@@ -490,22 +497,6 @@ def recursiveImport(modulename, baseDir=None, noCWD=0, debug=0):
     finally:
         sys.path = opath
 
-def recursiveGetAttr(obj, name):
-    "Can call down into e.g. object1.object2[4].attr"
-    return eval(name, obj.__dict__)
-
-def recursiveSetAttr(obj, name, value):
-    "Can call down into e.g. object1.object2[4].attr = value"
-    #get the thing above last.
-    tokens = name.split('.')
-    if len(tokens) == 1:
-        setattr(obj, name, value)
-    else:
-        most = '.'.join(tokens[:-1])
-        last = tokens[-1]
-        parent = recursiveGetAttr(obj, most)
-        setattr(parent, last, value)
-
 def import_zlib():
     try:
         import zlib
@@ -558,9 +549,9 @@ def getArgvDict(**kw):
             elif isinstance(v,int):
                 v = int(av)
             elif isinstance(v,list):
-                v = list(eval(av))
+                v = list(literal_eval(av),{})
             elif isinstance(v,tuple):
-                v = tuple(eval(av))
+                v = tuple(literal_eval(av),{})
             else:
                 raise TypeError("Can't convert string %r to %s" % (av,type(v)))
         return v
@@ -575,7 +566,7 @@ def getArgvDict(**kw):
         handled = 0
         ke = k+'='
         for a in A:
-            if a.find(ke)==0:
+            if a.startswith(ke):
                 av = a[len(ke):]
                 A.remove(a)
                 R[k] = handleValue(v,av,func)
@@ -714,6 +705,25 @@ def rl_getmtime(pn,os_path_isfile=os.path.isfile,os_path_normpath=os.path.normpa
     d = e[6]
     return time_mktime((((d>>9)&0x7f)+1980,(d>>5)&0xf,d&0x1f,(s>>11)&0x1f,(s>>5)&0x3f,(s&0x1f)<<1,0,0,0))
 
+if isPy3 and sys.version_info[:2]>=(3,4):
+    from importlib import util as importlib_util
+    def __rl_get_module__(name,dir):
+        for ext in ('.py','.pyw','.pyo','.pyc','.pyd'):
+            path = os.path.join(dir,name+ext)
+            if os.path.isfile(path):
+                spec = importlib_util.spec_from_file_location(name,path)
+                return spec.loader.load_module()
+        raise ImportError('no suitable file found')
+else:
+    import imp
+    def __rl_get_module__(name,dir):
+        f, p, desc= imp.find_module(name,[dir])
+        try:
+            return imp.load_module(name,f,p,desc)
+        finally:
+            if f:
+                f.close()
+
 def rl_get_module(name,dir):
     if name in sys.modules:
         om = sys.modules[name]
@@ -721,10 +731,8 @@ def rl_get_module(name,dir):
     else:
         om = None
     try:
-        f = None
         try:
-            f, p, desc= imp.find_module(name,[dir])
-            return imp.load_module(name,f,p,desc)
+            return __rl_get_module__(name,dir)
         except:
             if isCompactDistro():
                 #attempt a load from inside the zip archive
@@ -736,8 +744,6 @@ def rl_get_module(name,dir):
             raise ImportError('%s[%s]' % (name,dir))
     finally:
         if om: sys.modules[name] = om
-        del om
-        if f: f.close()
 
 def _isPILImage(im):
     try:
@@ -867,11 +873,12 @@ class ImageReader(object):
                 else:
                     im = self._image
                     mode = self.mode = im.mode
-                    if mode=='RGBA':
-                        if Image.VERSION.startswith('1.1.7'): im.load()
-                        self._dataA = ImageReader(im.split()[3])
-                        im = im.convert('RGB')
-                        self.mode = 'RGB'
+                    if mode in ('LA','RGBA'):
+                        if getattr(Image,'VERSION','').startswith('1.1.7'): im.load()
+                        self._dataA = ImageReader(im.split()[3 if mode=='RGBA' else 1])
+                        nm = mode[:-1]
+                        im = im.convert(nm)
+                        self.mode = nm
                     elif mode not in ('L','RGB','CMYK'):
                         if im.format=='PNG' and im.mode=='P' and 'transparency' in im.info:
                             im = im.convert('RGBA')
@@ -1446,3 +1453,88 @@ def makeFileName(s):
     if not isUnicode(s):
         s = s.decode('utf8')
     return s
+
+class FixedOffsetTZ(datetime.tzinfo):
+    """Fixed offset in minutes east from UTC."""
+
+    def __init__(self, h, m, name):
+        self.__offset = datetime.timedelta(hours=h, minutes = m)
+        self.__name = name
+
+    def utcoffset(self, dt):
+        return self.__offset
+
+    def tzname(self, dt):
+        return self.__name
+
+    def dst(self, dt):
+        return datetime.timedelta(0)
+
+class TimeStamp(object):
+    def __init__(self,invariant=None):
+        if invariant is None:
+            from reportlab.rl_config import invariant
+        t = os.environ.get('SOURCE_DATE_EPOCH','').strip()
+        if invariant or t:
+            t = int(t) if t else 946684800.0
+            lt = time.gmtime(t)
+            dhh = dmm = 0
+            self.tzname = 'UTC'
+        else:
+            t = time.time()
+            lt = tuple(time.localtime(t))
+            dhh = int(time.timezone / (3600.0))
+            dmm = (time.timezone % 3600) % 60
+            self.tzname = '' 
+        self.t = t
+        self.lt = lt
+        self.YMDhms = tuple(lt)[:6]
+        self.dhh = dhh
+        self.dmm = dmm
+
+    @property
+    def datetime(self):
+        if self.tzname:
+            return datetime.datetime.fromtimestamp(self.t,FixedOffsetTZ(self.dhh,self.dmm,self.tzname))
+        else:
+            return datetime.datetime.now()
+
+    @property
+    def asctime(self):
+        a = time.asctime(self.lt)
+        if not isPy3:
+            a = a.split()
+            a[2] = a[2].lstrip('0')
+            a = ' '.join(a)
+        return a
+
+def recursiveGetAttr(obj, name, g=None):
+    "Can call down into e.g. object1.object2[4].attr"
+    if not isStr(name): raise TypeError('invalid reursive acess using %r' % name)
+    name = asNative(name)
+    name = name.strip()
+    if not name: raise ValueError('empty recursive access')
+    dot = '.' if name and name[0] not in '[.(' else ''
+    return rl_safe_eval('obj%s%s'%(dot,name), g={}, l=dict(obj=obj))
+
+def recursiveSetAttr(obj, name, value):
+    "Can call down into e.g. object1.object2[4].attr = value"
+    #get the thing above last.
+    tokens = name.split('.')
+    if len(tokens) == 1:
+        setattr(obj, name, value)
+    else:
+        most = '.'.join(tokens[:-1])
+        last = tokens[-1]
+        parent = recursiveGetAttr(obj, most)
+        setattr(parent, last, value)
+
+def recursiveDelAttr(obj, name):
+    tokens = name.split('.')
+    if len(tokens) == 1:
+        delattr(obj, name)
+    else:
+        most = '.'.join(tokens[:-1])
+        last = tokens[-1]
+        parent = recursiveGetAttr(obj, most)
+        delattr(parent, last)
